﻿#include "pch.h"

#pragma comment(lib,"comctl32.lib")
#pragma comment(lib,"pdh.lib")
#pragma comment(lib,"imm32.lib")
#pragma comment(lib, "gdiplus")
#pragma comment(lib, "shlwapi.lib")
#pragma comment(lib, "winmm.lib")

// global.h에 선언하는 전역 변수
HINSTANCE g_hInst;
HWND hOptionPopup;
CMainWindow mainWindow;
CTrayWindow trayWindow;
CSetting setting(true);
CGdiPlusStarter g_gps;

// 이하의 전역 변수는 차후 옵션 구조체로 묶을 것
bool bVertical;
int bdStyle;			// 1:단색, 2:양각
int bdThick;			// 두께
COLORREF bdColor;		// 색상
int capThick;
int splitThick;
COLORREF splitColor;
int alwaysOnTop;
int widgetMinThick;		// 위젯 최소, 최대 크기
int widgetMaxThick;
int vertThick, horzThick;		// 각 방향별 두께. 경계선 포함. 사용자가 조정해야 바뀜
bool magneticEffet;				// 자석 효과 사용 여부

// 런타임 데이터
HFONT mainFont;
HFONT smallFont;
HFONT bigFont;
int tidgetSize;
int tidgetGap;
int clientWidth, clientHeight;		// 방향에 따라 바뀌는 현재 폭과 높이. 경계선 포함
int trayThick;
bool modiSetting;					// 전역 세팅 변경 여부
bool modiDget;						// 위젯, 티젯 편집 여부.

// 모듈 전역 변수 -> CMain의 멤버로 선언해도 됨
HWND hBtnClose, hBtnMenu;
HCURSOR hCurHorz, hCurVert;
int dragWidget;

// 위젯, 티젯 하나에 대한 정보.
struct sDgetInfo
{
	bool isWidget;				// true이면 위젯, false이면 티젯
	bool allowMulti;			// 멀티 인스턴스 가능 여부. 핫키, 리소스 과다 사용 등의 문제가 없으면 웬만하면 허용
	TCHAR name[32];				// 문자열로 된 위젯 설명. UI에 사용
	TCHAR initId[32];			// 위젯의 고유 명칭 접두어.멀티인 경우 일련 번호가 붙는다.
	int initHeight;				// 수직 배치일 때의 초기 높이. 최초 배치시, 리셋시 적용되며 사용자 편집 가능
	int initWidth;				// 수평 배치일 때의 초기 너비
	COLORREF initColor;			// 배경색
};

// 이 프로그램이 보유한 위젯, 티젯 목록. 이 중 사용자가 선택한 것만 사용한다. 
sDgetInfo arDget[] = {
	{true, true, L"시계", L"Clock", 80, 120, GetColorFromName(L"PapayaWhip")},
	{true, true, L"CPU 사용량", L"CpuUsage", 100, 140, GetColorFromName(L"PaleTurquoise")},
	{true, true, L"메모리 사용량", L"MemUsage", 50, 120, GetColorFromName(L"LightBlue")},
	{true, true, L"배터리", L"Battery", 50, 120, GetColorFromName(L"LightSkyBlue")},
	{true, true, L"디스크 사용량", L"DiskUsage", 50, 120, GetColorFromName(L"Wheat")},
	{true, false, L"음악 재생기", L"MusicPlayer", 150, 120, GetColorFromName(L"Gainsboro")},
	{true, true, L"메모", L"Memo", 150, 150, GetColorFromName(L"Ivory")},
	{true, true, L"계산기", L"Calc", 50, 120, GetColorFromName(L"Pink")},
	{true, true, L"진법변환기", L"Radix", 50, 120, GetColorFromName(L"Violet")},
	{true, true, L"이미지슬라이더", L"ImageSlide", 100, 150, GetColorFromName(L"DarkGray")},

	{false, false, L"LCD 테스트", L"LCDTest", 0x000000},
	{false, false, L"크기 측정", L"SizeMeter", 0x000000},
	{false, false, L"커서 이동", L"MoveCursor", 0x000000},
	{false, false, L"클립보드 변환", L"TextClip", 0x000000},
	{false, true, L"참조 이미지", L"OneImage", 0x606060},
	{false, false, L"테트리스", L"Tetris", 0x808080},
	{false, false, L"색상표", L"ColorTable", 0x808080},
};

// 표시할 위젯의 정보. 
struct sWidgetInfo
{
	TCHAR name[32];				// 위젯의 명칭
	TCHAR initId[32];			// 타입 구분을 위한 ID
	TCHAR id[32];				// 인스턴스의 고유 ID
	bool visible;				// 정보는 유지한채로 잠시 숨김
	int height;					// 수직일 때의 높이
	int width;					// 수평일 때의 너비
	COLORREF backColor;			// 배경 색상. 이 색으로 브러시를 만들어 둔다. 
	bool span;					// true이면 앞 위젯과 셀 공유
	int spanRatio;				// 분할량. 양수는 비율, 음수는 픽셀

	// 런타임 데이터
	CWidget* pWidget;			// 위젯 객체
};

// 위젯의 배열. 이 배열 내용대로 표시한다. 사용자가 편집할 수 있으며 저장, 복구 대상
TDArray<sWidgetInfo> arWidget;

// 위젯의 정보를 하나의 문자열로 합친다.
void PackWidgetInfo(int wi, TCHAR pack[])
{
	wsprintf(pack, L"%s§%s§%s§%d§%d§%d§%d§%d§%d",
		arWidget[wi].name, arWidget[wi].initId, arWidget[wi].id, arWidget[wi].visible,
		arWidget[wi].height, arWidget[wi].width, arWidget[wi].backColor, arWidget[wi].span,
		arWidget[wi].spanRatio);
}

// §마크로 구분된 문자열에서 idx번째 토큰을 읽는다. 
bool GetPackToken(LPCTSTR pack, int idx, TCHAR token[MAX_PATH])
{
	LPCTSTR p = pack;
	LPCTSTR pEnd;
	// idx 번째 §마크를 찾는다.
	for (int i = 0;;i++) {
		if (i == idx) break;
		pEnd = _tcsstr(p, L"§");
		if (pEnd == nullptr) return false;
		p = pEnd + 1;
	}

	// p부터 다음 § 또는 문자열 끝까지 읽는다.
	memset(token, 0, sizeof(TCHAR) * MAX_PATH);
	pEnd = _tcsstr(p, L"§");
	if (pEnd != nullptr) {
		_tcsncpy(token, p, pEnd - p);
	} else {
		lstrcpy(token, p);
	}

	return true;
}

sWidgetInfo UnpackWidgetInfo(LPCTSTR pack)
{
	TCHAR token[MAX_PATH];
	sWidgetInfo widget;
	GetPackToken(pack, 0, token);lstrcpy(widget.name, token);
	GetPackToken(pack, 1, token); lstrcpy(widget.initId, token);
	GetPackToken(pack, 2, token); lstrcpy(widget.id, token);
	GetPackToken(pack, 3, token); widget.visible = (token[0] == '1');
	GetPackToken(pack, 4, token); widget.height = _ttoi(token);
	GetPackToken(pack, 5, token); widget.width = _ttoi(token);
	GetPackToken(pack, 6, token); widget.backColor = _ttoi(token);
	GetPackToken(pack, 7, token); widget.span = (token[0] == '1');
	GetPackToken(pack, 8, token); widget.spanRatio = _ttoi(token);

	widget.pWidget = nullptr;
	return widget;
}

// 이름으로부터 위젯 초기 정보를 생성한다.
// 최초 설치시, 위젯 추가시 호출한다.
void CopyWidgetInfo(LPCTSTR initId, sWidgetInfo* widget)
{
	int idx;
	for (idx = 0; idx < Count(arDget);idx++) {
		if (lstrcmp(arDget[idx].initId, initId) == 0) break;
	}

	lstrcpy(widget->name, arDget[idx].name);
	lstrcpy(widget->initId, arDget[idx].initId);
	if (arDget[idx].allowMulti == false) {
		lstrcpy(widget->id, arDget[idx].initId);
	} else {
		bool bFind;
		TCHAR id[32];
		// 일련 번호를 증가시키 가면 빈 번호를 찾는다. 
		for (int num = 1;; num++) {
			wsprintf(id, L"%s_%d", initId, num);
			bFind = false;
			for (int wi = 0; wi < arWidget.num; wi++) {
				if (lstrcmp(arWidget[wi].id, id) == 0) {
					bFind = true;
					break;
				}
			}
			if (bFind == false) break;
		}

		lstrcpy(widget->id, id);
	}

	widget->visible = true;
	widget->height = arDget[idx].initHeight;
	widget->width = arDget[idx].initWidth;
	widget->backColor = arDget[idx].initColor;
	widget->span = false;
	widget->spanRatio = 100;

	widget->pWidget = nullptr;
}

// 이름으로부터 위젯 객체 생성. 타입 정보를 포함할 수 없어 직접 생성한다.
CWidget* CreateWidgetObject(LPCTSTR initId, LPCTSTR name, LPCTSTR id)
{
	if (lstrcmp(initId, L"Clock") == 0) return new CClock(name, id);
	if (lstrcmp(initId, L"CpuUsage") == 0) return new CCpuUsage(name, id);
	if (lstrcmp(initId, L"MemUsage") == 0) return new CMemUsage(name, id);
	if (lstrcmp(initId, L"DiskUsage") == 0) return new CDiskUsage(name, id);
	if (lstrcmp(initId, L"Battery") == 0) return new CBattery(name, id);
	if (lstrcmp(initId, L"MusicPlayer") == 0) return new CMusicPlayer(name, id);
	if (lstrcmp(initId, L"Memo") == 0) return new CMemo(name, id);
	if (lstrcmp(initId, L"Calc") == 0) return new CCalc(name, id);
	if (lstrcmp(initId, L"Radix") == 0) return new CRadix(name, id);
	if (lstrcmp(initId, L"ImageSlide") == 0) return new CImageSlide(name, id);
	return NULL;
}

// arWidget 배열의 wi번째 정보를 참조하여 위젯을 생성하고 초기화한다. 
void CreateWidget(int wi, HWND hParent)
{
	// 위젯 객체 생성
	arWidget[wi].pWidget = CreateWidgetObject(arWidget[wi].initId, arWidget[wi].name, arWidget[wi].id);

	// 높이와 너비 지정. x, y는 Relayout에서 계산한다. 
	arWidget[wi].pWidget->win_h = arWidget[wi].height;
	arWidget[wi].pWidget->win_w = arWidget[wi].width;

	// 현재 색상으로 배경 브러시 생성
	arWidget[wi].pWidget->win_backBrush = CreateSolidBrush(arWidget[wi].backColor);

	// 위젯 윈도우 생성
	arWidget[wi].pWidget->Create(hParent);
}

// 표시할 티젯의 정보
struct sTidgetInfo
{
	TCHAR name[32];				// 티젯의 명칭
	TCHAR initId[32];			// 타입 구분을 위한 ID
	TCHAR id[32];				// 인스턴스의 고유 ID
	bool visible;				// 정보는 유지한채로 잠시 숨김

	CTidget* pTidget;
};

// 티젯의 배열
TDArray<sTidgetInfo> arTidget;

void PackTidgetInfo(int wi, TCHAR pack[])
{
	wsprintf(pack, L"%s§%s§%s§%d",
		arTidget[wi].name, arTidget[wi].initId, arTidget[wi].id, arTidget[wi].visible);
}

sTidgetInfo UnpackTidgetInfo(LPCTSTR pack)
{
	TCHAR token[MAX_PATH];
	sTidgetInfo tidget;
	GetPackToken(pack, 0, token); lstrcpy(tidget.name, token);
	GetPackToken(pack, 1, token); lstrcpy(tidget.initId, token);
	GetPackToken(pack, 2, token); lstrcpy(tidget.id, token);
	GetPackToken(pack, 3, token); tidget.visible = (token[0] == '1');

	tidget.pTidget = nullptr;
	return tidget;
}

// 이름으로부터 위젯 초기 정보를 생성한다.
void CopyTidgetInfo(LPCTSTR initId, sTidgetInfo* tidget)
{
	int idx;
	for (idx = 0; idx < Count(arDget); idx++) {
		if (lstrcmp(arDget[idx].initId, initId) == 0) break;
	}

	lstrcpy(tidget->name, arDget[idx].name);
	lstrcpy(tidget->initId, arDget[idx].initId);
	if (arDget[idx].allowMulti == false) {
		lstrcpy(tidget->id, arDget[idx].initId);
	} else {
		// 멀티인 경우는 기존 티젯과 중복이 없는 id를 찾는다. 
		bool bFind;
		TCHAR id[32];
		for (int num = 1;; num++) {
			wsprintf(id, L"%s_%d", initId, num);
			bFind = false;
			for (int ti = 0; ti < arTidget.num; ti++) {
				if (lstrcmp(arTidget[ti].id, id) == 0) {
					bFind = true;
					break;
				}
			}
			if (bFind == false) break;
		}

		lstrcpy(tidget->id, id);
	}
	tidget->visible = true;

	tidget->pTidget = nullptr;
}

CTidget* CreateTidgetObject(LPCTSTR initId, LPCTSTR name, LPCTSTR id)
{
	if (lstrcmp(initId, L"LCDTest") == 0) return new CLCDTest(name, id);
	if (lstrcmp(initId, L"SizeMeter") == 0) return new CSizeMeter(name, id);
	if (lstrcmp(initId, L"OneImage") == 0) return new COneImage(name, id);
	if (lstrcmp(initId, L"MoveCursor") == 0) return new CMoveCursor(name, id);
	if (lstrcmp(initId, L"TextClip") == 0) return new CTextClip(name, id);
	if (lstrcmp(initId, L"Tetris") == 0) return new CTetris(name, id);
	if (lstrcmp(initId, L"ColorTable") == 0) return new CColorTable(name, id);
	return NULL;
}

// arTidget 배열의 ti번째 정보를 참조하여 위젯을 생성하고 초기화한다. 
void CreateTidget(int ti, HWND hParent)
{
	// 객체 생성
	arTidget[ti].pTidget = CreateTidgetObject(arTidget[ti].initId, arTidget[ti].name, arTidget[ti].id);

	// 윈도우 생성
	arTidget[ti].pTidget->Create(hParent);
}


int APIENTRY WinMain(_In_ HINSTANCE hInstance, _In_opt_ HINSTANCE hPrevInstance
	, _In_ LPSTR lpszCmdParam, _In_ int nCmdShow)
{
	MSG Message;
	g_hInst = hInstance;
	InitCommonControls();
	srand(GetTickCount());
	TCHAR key[32];
	TCHAR pack[256];

	// ini 파일 경로 초기화. 실행 파일명에 상관 없이 초기화 파일명은 고정한다.
	//setting.SetIniPath(L"SoEnty.ini");
	// 실행 파일명을 따라 가도록 한다. 그래야 SoEnty100.exe, SoEnty=8월1일자.exe 식으로 버전별로 이름 바꿔도 각 버전 상태를 유지할 수 있다.
	// 또 멀티 인스턴스로 실행할 수 있어 이름 바꿔 놓고 왼쪽 하나, 오른쪽 하나 붙일 수 있다. 
	setting.SetIniPath(NULL);
	lstrcpy(setting.regRoot, L"Software\\SoEnSoft\\SoEnty\\");
	// 레지스트리에 기록할 때는 이 주석을 풀면 된다.
	//setting.useIni = false;

	// 전역 변수 초기화.
	bVertical = setting.get(L"main", L"vertical", 1); 
	alwaysOnTop = setting.get(L"main", L"alwaysontop", 1); 
	widgetMinThick = setting.get(L"main", L"widgetMinThick", 30);
	widgetMaxThick = setting.get(L"main", L"widgetMaxThick", 300);
	bdStyle = setting.get(L"main", L"bdStyle", 2); 
	bdThick = setting.get(L"main", L"bdThick", 4);
	bdColor = setting.get(L"main", L"bdColor", (int)RGB(192, 192, 192));
	capThick = setting.get(L"main", L"capThick", 24);
	splitThick = setting.get(L"main", L"splitThick", 5);
	splitColor = setting.get(L"main", L"splitColor", (int)GetColorFromName(L"indigo"));
	vertThick = setting.get(L"main", L"vertThick", 200);
	horzThick = setting.get(L"main", L"horzThick", 120);
	tidgetSize = setting.get(L"main", L"tidgetSize", 24);
	tidgetGap = setting.get(L"main", L"tidgetGap", 4);
	magneticEffet = setting.get(L"main", L"magneticEffet", 1); 

	// 위젯 목록 읽기
	int wnum = setting.get(L"main", L"wnum", 0);
	int tnum = setting.get(L"main", L"tnum", 0);
	sWidgetInfo widget;
	sTidgetInfo tidget;
	if (wnum == 0) {
		// 첫 실행시 초기 표시할 위젯의 정보. 최초 설치시의 초기 배치 상태이다.
		CopyWidgetInfo(L"Clock", &widget);arWidget.Append(widget);
		// 두 개도 배치 가능하다. 
		// CopyWidgetInfo(L"Clock", &widget); arWidget.Append(widget);
		CopyWidgetInfo(L"CpuUsage", &widget);arWidget.Append(widget);
		CopyWidgetInfo(L"MemUsage", &widget);arWidget.Append(widget);
		// 배터리는 메모리와 합침
		CopyWidgetInfo(L"Battery", &widget); widget.span = true; arWidget.Append(widget);
		CopyWidgetInfo(L"DiskUsage", &widget); arWidget.Append(widget);
		CopyWidgetInfo(L"MusicPlayer", &widget); arWidget.Append(widget);
		CopyWidgetInfo(L"Memo", &widget); arWidget.Append(widget);
		CopyWidgetInfo(L"Calc", &widget); arWidget.Append(widget);
		CopyWidgetInfo(L"Radix", &widget); arWidget.Append(widget);
		CopyWidgetInfo(L"ImageSlide", &widget); arWidget.Append(widget);

		CopyTidgetInfo(L"LCDTest", &tidget);arTidget.Append(tidget);
		CopyTidgetInfo(L"SizeMeter", &tidget); arTidget.Append(tidget);
		CopyTidgetInfo(L"OneImage", &tidget);arTidget.Append(tidget);
		// 이미지 하나 더 배치. 각각 다른 이미지 정보 가질 수 있다.
		// CopyTidgetInfo(L"OneImage", &tidget); arTidget.Append(tidget);
		CopyTidgetInfo(L"MoveCursor", &tidget); arTidget.Append(tidget);
		CopyTidgetInfo(L"TextClip", &tidget); arTidget.Append(tidget);
		CopyTidgetInfo(L"Tetris", &tidget); arTidget.Append(tidget);
		CopyTidgetInfo(L"ColorTable", &tidget); arTidget.Append(tidget);
		modiDget = true;
	} else {
		// INI 파일에 저장된 위젯 목록과 속성을 읽어온다. 
		for (int wi = 0; wi < wnum; wi++) {
			wsprintf(key, L"widget_%d", wi);
			setting.get(L"main", key, L"", pack, 256);
			widget = UnpackWidgetInfo(pack);
			arWidget.Append(widget);
		}

		for (int ti = 0; ti < tnum; ti++) {
			wsprintf(key, L"tidget_%d", ti);
			setting.get(L"main", key, L"", pack, 256);
			tidget = UnpackTidgetInfo(pack);
			arTidget.Append(tidget);
		}
		modiDget = false;
	}

	// 메인 윈도우 생성
	mainWindow.Create(alwaysOnTop);
	ShowWindow(mainWindow.win_hWnd, nCmdShow);

	while (GetMessage(&Message, NULL, 0, 0)) {
		if (!IsWindow(hOptionPopup) || !IsDialogMessage(hOptionPopup, &Message)) {
			TranslateMessage(&Message);
			DispatchMessage(&Message);
		}
	}
	return (int)Message.wParam;
}

LRESULT CMainWindow::OnMessage(UINT iMessage, WPARAM wParam, LPARAM lParam) {
	POINT pt;

	switch (iMessage)
	{
	case WM_SETCURSOR:
		GetCursorPos(&pt);
		ScreenToClient(win_hWnd, &pt);

		if (LOWORD(lParam) == HTCLIENT) {
			if (bVertical) {
				if (FindSplitVirt(pt.y) != -1) {
					SetCursor(hCurVert);
					return 0;
				}
			} else {
				if (FindSplitHorz(pt.x) != -1) {
					SetCursor(hCurHorz);
					return 0;
				}
			}
		}
		break;
		// 두께는 최소 100 ~ 640까지
	case WM_GETMINMAXINFO:
		LPMINMAXINFO pmmi;
		pmmi = (LPMINMAXINFO)lParam;
		if (bVertical) {
			pmmi->ptMinTrackSize.x = 100;
			pmmi->ptMaxTrackSize.x = 640;
		} else {
			pmmi->ptMinTrackSize.y = 100;
			pmmi->ptMaxTrackSize.y = 640;
		}
		return 0;
	case WM_NCHITTEST:
		int nHit;
		nHit = (int)DefWindowProc(win_hWnd, WM_NCHITTEST, wParam, lParam);
		// 음수 좌표가 전달될 수도 있어 부호 변환 캐스팅이 반드시 필요하다.
		pt.x = (int)(short)LOWORD(lParam);
		pt.y = (int)(short)HIWORD(lParam);
		ScreenToClient(win_hWnd, &pt);
		RECT crt;
		GetClientRect(win_hWnd, &crt);

		if (nHit == HTCLIENT) {
			if (bVertical) {
				if (pt.x < bdThick)
					return HTLEFT;
				if (pt.x > crt.right - 1 - bdThick)
					return HTRIGHT;
			} else {
				if (pt.y < bdThick)
					return HTTOP;
				if (pt.y > crt.bottom - 1 - bdThick)
					return HTBOTTOM;
			}
			if (IsInCaption(pt.x, pt.y)) {
				return HTCAPTION;
			}
		}

		return nHit;
	case WM_WINDOWPOSCHANGING:
		if (magneticEffet) {
			LPWINDOWPOS wp = (LPWINDOWPOS)lParam;
			HMONITOR hMonitor;
			MONITORINFOEX mi;
			hMonitor = MonitorFromWindow(mainWindow.win_hWnd, MONITOR_DEFAULTTONEAREST);
			mi.cbSize = sizeof(MONITORINFOEX);
			GetMonitorInfo(hMonitor, &mi);
			RECT mrt;
			// 현재 모니터의 화면 영역이 아닌 작업 영역을 기준으로 한다. 
			CopyRect(&mrt, &mi.rcWork);

			// 모니터의 벽에 달라 붙는다. 
			if (ABS(wp->x - mrt.left) < 30) wp->x = mrt.left;
			if (ABS(wp->x + wp->cx - mrt.right) < 30) wp->x = mrt.right - wp->cx;
			if (ABS(wp->y - mrt.top) < 30) wp->y = mrt.top;
			if (ABS(wp->y + wp->cy - mrt.bottom) < 30) wp->y = mrt.bottom - wp->cy;
		}
		return 0;
	case WM_MOVE:
		if (IsWindowVisible(win_hWnd)) {
			modiSetting = true;
		}
		return 0;
	}

	return CWindow::OnMessage(iMessage, wParam, lParam);
}

void CMainWindow::Create(bool alwaysOnTop) 
{
	// 최후 실행 위치를 계산한다. OnCreate에서 해도 되지만 아무 곳에나 배치해 두면 
	// 처음 뜰때 깜박임이 있어 생성할 때부터 최후 위치로 대충 잡았다. 
	RECT wrt;
	wrt.left = setting.get(L"main", L"left", 100);
	wrt.top = setting.get(L"main", L"top", 100);
	wrt.right = setting.get(L"main", L"right", 300);
	wrt.bottom = setting.get(L"main", L"bottom", 600);

	// todo : INI 파일에서 읽은 위치가 현재 모니터 구성상 보이는 곳인지 점검 후 조정.

	// 차일드 영역은 그리기에서 제외한다. 
	win_hWnd = CreateWindowEx(alwaysOnTop ? WS_EX_TOPMOST : 0,
		TEXT("CWindow"), L"SoEnty", WS_POPUP | WS_CLIPCHILDREN,
		wrt.left, wrt.top, wrt.right - wrt.left, wrt.bottom - wrt.top,
		NULL, (HMENU)NULL, g_hInst, this);
	HICON hIcon = LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_SOENTY));
	SetClassLong(win_hWnd, GCL_HICON, (LONG)hIcon);
}

LRESULT CMainWindow::OnCreate(WPARAM wParam, LPARAM lParam)
{
	// 닫기 버튼 - 릴리즈에서는 팝업 메뉴 안에 숨기는게 어떨지. 디버깅 편의를 위해 필요함. 
	hBtnClose = CreateWindow(TEXT("GrpBtn"), NULL, WS_CHILD | WS_VISIBLE | GBS_PUSH,
		0, 0, 16, 16, win_hWnd, (HMENU)1, g_hInst, NULL);
	SendMessage(hBtnClose, GBM_SETIMAGE, MAKEWPARAM(IDB_CLOSE_NORMAL, IDB_CLOSE_HOT),
		MAKEWPARAM(IDB_CLOSE_DOWN, IDB_CLOSE_NORMAL));
	hBtnMenu = CreateWindow(TEXT("GrpBtn"), NULL, WS_CHILD | WS_VISIBLE | GBS_PUSH,
		0, 0, 16, 16, win_hWnd, (HMENU)2, g_hInst, NULL);
	SendMessage(hBtnMenu, GBM_SETIMAGE, MAKEWPARAM(IDB_MENU_NOR, IDB_MENU_NOR),
		MAKEWPARAM(IDB_MENU_NOR, IDB_MENU_NOR));

	hCurHorz = LoadCursor(g_hInst, MAKEINTRESOURCE(IDC_SIZEHORZ));
	hCurVert = LoadCursor(g_hInst, MAKEINTRESOURCE(IDC_SIZEVERT));

	mainFont = CreateFont(15, 0, 0, 0, 0, 0, 0, 0, HANGUL_CHARSET, 3, 2, 1,
		VARIABLE_PITCH | FF_MODERN, TEXT("맑은 고딕"));
	smallFont = CreateFont(12, 0, 0, 0, 0, 0, 0, 0, HANGUL_CHARSET, 3, 2, 1,
		VARIABLE_PITCH | FF_MODERN, TEXT("맑은 고딕"));
	bigFont = CreateFont(28, 0, 0, 0, 0, 0, 0, 0, HANGUL_CHARSET, 3, 2, 1,
		VARIABLE_PITCH | FF_MODERN, TEXT("맑은 고딕"));

	// 위젯 생성. 숨겨진 위젯도 생성 대상이다.
	for (int w = 0; w < arWidget.num; w++) {
		CreateWidget(w, win_hWnd);
	}

	// 트레이 생성
	trayWindow.Create();

	// 티젯 생성
	for (int t = 0; t < arTidget.num; t++) {
		CreateTidget(t, trayWindow.win_hWnd);
	}

	Relayout();

	// 초기 위치를 잡는다. 위치는 INI에서 읽은 값을 그대로 쓰고 크기만 Relayout에서 계산한 값을 사용한다. 
	SetWindowPos(win_hWnd, NULL, 0, 0, clientWidth, clientHeight, SWP_NOMOVE | SWP_NOZORDER);

	// 설정은 아직 변경하지 않은 것으로 기록한다. 단, 최초 실행시 Dget이 초기화된 경우는 설정도 무조건 저장한다.
	modiSetting = false;
	if (modiDget) {
		modiSetting = true;
	}
	return 0;
}

LRESULT CMainWindow::OnDestroy(WPARAM wParam, LPARAM lParam)
{
	// 옵션, 크기, 위치 등의 설정이 변경된 경우만 저장한다. 
	if (modiSetting) {
		RECT wrt;
		GetWindowRect(win_hWnd, &wrt);
		setting.set(L"main", L"left", (int)wrt.left);
		setting.set(L"main", L"top", (int)wrt.top);
		setting.set(L"main", L"right", (int)wrt.right);
		setting.set(L"main", L"bottom", (int)wrt.bottom);

		setting.set(L"main", L"vertical", (int)bVertical);
		setting.set(L"main", L"alwaysontop", (int)alwaysOnTop);
		setting.set(L"main", L"widgetMinThick", widgetMinThick);
		setting.set(L"main", L"widgetMaxThick", widgetMaxThick);
		setting.set(L"main", L"vertThick", vertThick);
		setting.set(L"main", L"horzThick", horzThick);
	}

	// 위젯, 티젯 배열을 저장한다. 저장 최소화를 위해 변경된 것이 있을 때만 저장한다. 
	if (modiDget) {
		TCHAR key[32];
		TCHAR pack[256];
		setting.set(L"main", L"wnum", arWidget.num);
		for (int wi = 0; wi < arWidget.num; wi++) {
			wsprintf(key, L"widget_%d", wi);
			PackWidgetInfo(wi, pack);
			setting.set(L"main", key, pack);
		}
		setting.set(L"main", L"tnum", arTidget.num);
		for (int ti = 0; ti < arTidget.num; ti++) {
			wsprintf(key, L"tidget_%d", ti);
			PackTidgetInfo(ti, pack);
			setting.set(L"main", key, pack);
		}
	}

	// 메모 : 주요 전역 변수만 저장하고 잘 바뀌지 않는 개별 설정값을 각 젯이 알아서 저장한다.

	DeleteObject(mainFont);
	DeleteObject(smallFont);
	PostQuitMessage(0);
	return 0;
}

LRESULT CMainWindow::OnCommand(WPARAM wParam, LPARAM lParam)
{
	switch (LOWORD(wParam)) {
	case 1:
		DestroyWindow(win_hWnd);
		break;
	case 2:
		HMENU hMenu, hPopup;
		hMenu = LoadMenu(g_hInst, MAKEINTRESOURCE(IDR_POPUPMENU));
		hPopup = GetSubMenu(hMenu, 0);

		if (IsTopMost(win_hWnd)) {
			CheckMenuItem(hPopup, ID_ALWAYSONTOP, MF_BYCOMMAND | MF_CHECKED);
		}

		POINT pt;
		GetCursorPos(&pt);
		TrackPopupMenu(hPopup, TPM_LEFTALIGN, pt.x, pt.y, 0, win_hWnd, NULL);
		DestroyMenu(hMenu);
		break;
	case ID_ROTATE:
		modiSetting = true;
		// 임시로 방향 스위칭 기능으로 사용한다. 차후 팝업 메뉴 안에 넣을 예정이다.
		bVertical = !bVertical;
		Relayout();

		// 메인 윈도우 크기 변경.
		// todo : 방향 전환에 의해 일부가 화면을 벗어나면 안쪽으로 넣어주기
		SetWindowPos(win_hWnd, NULL, 0, 0, clientWidth, clientHeight, SWP_NOMOVE | SWP_NOZORDER);
		break;
	case ID_ALWAYSONTOP:
		modiSetting = true;
		if (IsTopMost(win_hWnd)) {
			SetWindowPos(win_hWnd, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
			alwaysOnTop = false;
		} else {
			SetWindowPos(win_hWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
			alwaysOnTop = true;
		}
		break;
	case ID_ABOUT:
		MessageBox(win_hWnd, L"여러 가지 간단한 위젯을 하나로 묶어 놓은 패키지 프로그램이다.\r\n"
			L"누구나 무료로 쓸 수 있되 맘에 들면 기부할 수 있는 도네이션웨어이다.\r\n"
			L"(단, 조기봉, 조종권, 박요한은 일단 기부부터 해야 함)",
			L"SoEnty", MB_OK);
		return 0;
	case ID_SETTING:
		modiSetting = true;
		MessageBox(win_hWnd, L"설정 대화상자 아직 미작성\r\n",
			L"SoEnty", MB_OK);
		break;
	case ID_HELP:
		MessageBox(win_hWnd, L"도움말은 차후 기능이 어느 정도 완성된 후에 제공할 예정입니다.\r\n",
			L"SoEnty", MB_OK);
		break;
	case ID_SELWIDGET:
		modiDget = true;
		DialogBox(g_hInst, MAKEINTRESOURCE(IDD_SELWIDGET), win_hWnd, SelWidgetDlgProc);
		break;
	case ID_SELTIDGET:
		modiDget = true;
		DialogBox(g_hInst, MAKEINTRESOURCE(IDD_SELTIDGET), win_hWnd, SelTidgetDlgProc);
		break;
	}

	return 0;
}

// 메인 윈도우는 차일드를 덮어쓰는 문제가 있어 더블버퍼링하지 않는다. 
LRESULT CMainWindow:: OnPaint(WPARAM wParam, LPARAM lParam)
{
	HDC hdc;
	PAINTSTRUCT ps;
	HPEN light, dark, oldPen;
	HBRUSH bdBrush, splitBrush;
	LPCTSTR caption = L"SoEnty";

	RECT crt;
	GetClientRect(win_hWnd, &crt);
	hdc = BeginPaint(win_hWnd, &ps);

	// 경계선을 그린다. 
	bdBrush = CreateSolidBrush(bdColor);
	FillRect(hdc, 0, 0, clientWidth, bdThick, bdBrush);
	FillRect(hdc, 0, 0, bdThick, clientHeight, bdBrush);
	FillRect(hdc, 0, clientHeight - bdThick, clientWidth, clientHeight, bdBrush);
	FillRect(hdc, clientWidth - bdThick, 0, clientWidth, clientHeight, bdBrush);
	splitBrush = CreateSolidBrush(splitColor);

	// 위쪽에 타이틀 바를 그린다.
	HFONT font, oldfont;
	SetBkMode(hdc, TRANSPARENT);

	// 경계선을 입맛대로 직접 그린다.
	if (bdStyle == 2) {
		light = CreatePen(PS_SOLID, 1, GetLightColor(bdColor, 64));
		dark = CreatePen(PS_SOLID, 1, GetDarkColor(bdColor, 64));

		oldPen = (HPEN)SelectObject(hdc, light);
		MoveToEx(hdc, 0, 0, NULL); LineTo(hdc, clientWidth, 0);
		MoveToEx(hdc, 0, 0, NULL); LineTo(hdc, 0, clientHeight);
		MoveToEx(hdc, clientWidth - bdThick, bdThick, NULL); LineTo(hdc, clientWidth - bdThick, clientHeight - bdThick);
		MoveToEx(hdc, bdThick, clientHeight - bdThick, NULL); LineTo(hdc, clientWidth - bdThick + 1, clientHeight - bdThick);
		SelectObject(hdc, dark);
		MoveToEx(hdc, bdThick, bdThick, NULL); LineTo(hdc, clientWidth - bdThick - 1, bdThick);
		MoveToEx(hdc, bdThick, bdThick, NULL); LineTo(hdc, bdThick, clientHeight - bdThick - 1);
		MoveToEx(hdc, clientWidth - 1, 0, NULL); LineTo(hdc, clientWidth - 1, clientHeight - 1);
		MoveToEx(hdc, 0, clientHeight - 1, NULL); LineTo(hdc, clientWidth, clientHeight - 1);

		SelectObject(hdc, oldPen);
		DeleteObject(light);
		DeleteObject(dark);
	}

	if (bVertical) {
		// 타이틀 영역 그림
		FillRect(hdc, bdThick, bdThick, clientWidth - bdThick, bdThick + capThick, GetSysColorBrush(COLOR_ACTIVECAPTION));

		font = CreateFont(15, 0, 0, 0, 0, 0, 0, 0, HANGUL_CHARSET, 3, 2, 1,
			VARIABLE_PITCH | FF_MODERN, TEXT("맑은 고딕"));
		oldfont = (HFONT)SelectObject(hdc, font);
		// 수직 중앙 위치에 타이틀 출력
		TextOutVCenter(hdc, bdThick + 4, bdThick + capThick / 2, caption);

		FillRect(hdc, bdThick, bdThick + capThick, clientWidth - bdThick, clientHeight - bdThick, splitBrush);

		/* 스팬 사이까지 채우기 번거로워 단색으로 채우는 코드로 변경함
		int y = bdThick + capThick;
		for (int i = 0; i < arWidget.num; i++) {
			if (arWidget[i].visible == false) continue;
			y += arWidget[i].pWidget->win_h;

			FillRect(hdc, bdThick, y, clientWidth - bdThick, y + splitThick, splitBrush);
			if (bdStyle == 2) {
				light = CreatePen(PS_SOLID, 1, GetLightColor(splitColor, 16));
				dark = CreatePen(PS_SOLID, 1, GetDarkColor(splitColor, 16));

				oldPen = (HPEN)SelectObject(hdc, light);
				MoveToEx(hdc, bdThick, y, NULL); LineTo(hdc, clientWidth - bdThick, y);
				SelectObject(hdc, dark);
				MoveToEx(hdc, bdThick, y + splitThick - 1, NULL); LineTo(hdc, clientWidth - bdThick, y + splitThick - 1);

				SelectObject(hdc, oldPen);
				DeleteObject(light);
				DeleteObject(dark);
			}

			y += splitThick;
		}*/
	} else {
		FillRect(hdc, bdThick, bdThick, bdThick + capThick, clientHeight - bdThick, GetSysColorBrush(COLOR_ACTIVECAPTION));

		font = CreateFont(15, 0, 900, 0, 0, 0, 0, 0, HANGUL_CHARSET, 3, 2, 1,
			VARIABLE_PITCH | FF_MODERN, TEXT("맑은 고딕"));
		oldfont = (HFONT)SelectObject(hdc, font);
		SIZE sz;
		GetTextExtentPoint32(hdc, caption, lstrlen(caption), &sz);
		TextOut(hdc, bdThick + capThick / 2 - sz.cy / 2, clientHeight - 50, caption, lstrlen(caption));

		FillRect(hdc, bdThick + capThick, bdThick, clientWidth - bdThick, clientHeight - bdThick, splitBrush);

		/*int x = bdThick + capThick;
		for (int i = 0; i < arWidget.num; i++) {
			if (arWidget[i].visible == false) continue;
			x += arWidget[i].pWidget->win_w;

			FillRect(hdc, x, bdThick, x + splitThick, clientHeight - bdThick, splitBrush);
			if (bdStyle == 2) {
				light = CreatePen(PS_SOLID, 1, GetLightColor(splitColor, 16));
				dark = CreatePen(PS_SOLID, 1, GetDarkColor(splitColor, 16));

				oldPen = (HPEN)SelectObject(hdc, light);
				MoveToEx(hdc, x, bdThick, NULL); LineTo(hdc, x, clientHeight - bdThick);
				SelectObject(hdc, dark);
				MoveToEx(hdc, x + splitThick - 1, bdThick, NULL); LineTo(hdc, x + splitThick - 1, clientHeight - bdThick);

				SelectObject(hdc, oldPen);
				DeleteObject(light);
				DeleteObject(dark);
			}

			x += splitThick;
		}*/
	}

	DeleteObject(bdBrush);
	SelectObject(hdc, oldfont);
	DeleteObject(font);
	EndPaint(win_hWnd, &ps);
	return 0;
}

LRESULT CMainWindow::OnKeyDown(WPARAM wParam, LPARAM lParam)
{
	if (wParam == 'A') {
		// 테스트 코드 작성용으로 사용할 것. 
	}
	return 0;
}

LRESULT CMainWindow::OnLButtonDown(WPARAM wParam, LPARAM lParam)
{
	POINT pt;

	pt.x = (int)(short)LOWORD(lParam);
	pt.y = (int)(short)HIWORD(lParam);

	if (bVertical) {
		dragWidget = FindSplitVirt(pt.y);
		if (dragWidget == -1) return 0;
		SetCapture(win_hWnd);
	} else {
		dragWidget = FindSplitHorz(pt.x);
		if (dragWidget == -1) return 0;
		SetCapture(win_hWnd);
	}

	return 0;
}

LRESULT CMainWindow::OnMouseMove(WPARAM wParam, LPARAM lParam)
{
	POINT pt;

	if (GetCapture() != win_hWnd) {
		return 0;
	}

	// 경계선을 조금이라도 드래그했으면 크기가 바뀐 것으로 본다. 
	modiDget = true;

	pt.x = (int)(short)LOWORD(lParam);
	pt.y = (int)(short)HIWORD(lParam);

	if (bVertical) {
		// 최소, 최대 높이 적용
		int newHeight = min(max(pt.y - arWidget[dragWidget].pWidget->win_y, widgetMinThick), widgetMaxThick);

		// 드래그중인 위젯의 높이 조정
		arWidget[dragWidget].pWidget->win_h = newHeight;
	} else {
		// 최소, 최대 폭 적용
		int newWidth = min(max(pt.x - arWidget[dragWidget].pWidget->win_x, widgetMinThick), widgetMaxThick);

		// 드래그중인 위젯의 높이 조정
		arWidget[dragWidget].pWidget->win_w = newWidth;
		CustomTrace(L"newWidth = %d - %d = %d\r\n", pt.x, arWidget[dragWidget].pWidget->win_x, newWidth);
	}

	// Relayout()을 호출할 필요 없이 메인 윈도우의 크기 변화만 계산한다.
	if (bVertical) {
		clientHeight = capThick + bdThick * 2;
		for (int i = 0; i < arWidget.num; i++) {
			if (arWidget[i].visible == false) continue;
			clientHeight += (arWidget[i].pWidget->win_h + splitThick);
		}
		clientHeight += trayThick;
	} else {
		clientWidth = capThick + bdThick * 2;
		for (int i = 0; i < arWidget.num; i++) {
			if (arWidget[i].visible == false) continue;
			clientWidth += (arWidget[i].pWidget->win_w + splitThick);
		}
		clientWidth += trayThick;
	}
	CustomTrace(L"Main.OnMouseMove. width = %d, height = %d, trayTich = %d\r\n", clientWidth, clientHeight, trayThick);

	SetWindowPos(win_hWnd, NULL, 0, 0, clientWidth, clientHeight, SWP_NOMOVE | SWP_NOZORDER);
	Relayout();
	// 메인 윈도우 즉시 다시 그림
	InvalidateRect(win_hWnd, NULL, FALSE);
	UpdateWindow(win_hWnd);
	return 0;
}

LRESULT CMainWindow::OnLButtonUp(WPARAM wParam, LPARAM lParam)
{
	dragWidget = -1;
	ReleaseCapture();
	return 0;
}

LRESULT CMainWindow::OnTimer(WPARAM wParam, LPARAM lParam)
{
	return 0;
}

LRESULT CMainWindow::OnSize(WPARAM wParam, LPARAM lParam)
{
	RECT crt;
	GetClientRect(win_hWnd, &crt);
	CustomTrace(L"Main.OnSize. width = %d, height = %d\r\n", crt.right, crt.bottom);

	// 크기가 조금이라도 바뀌면 설정 변경으로 기록한다. 단, 초기화중에 오는 메시지는 무시한다.
	// WM_MOVE 메시지도 마찬가지 방식으로 실제 윈도우가 이동할 때만 설정 변경으로 취급한다. 
	if (IsWindowVisible(win_hWnd)) {
		modiSetting = true;
	}

	// 폭 또는 높이를 변경한다.
	if (bVertical) {
		// 실제로 바뀐 경우에만 레이아웃을 다시 잡는다. Relayout에서 크기 변경시에도 WM_SIZE가 오는데 이때는
		// 크기가 이미 결정된 상태여서 다시 Relayout을 호출할 필요 없다. 
		if (vertThick != crt.right) {
			vertThick = crt.right;
			Relayout();
		}
	} else {
		if (horzThick != crt.bottom) {
			horzThick = crt.bottom;
			Relayout();
		}
	}
	return 0;
}

// wi 번째 이후의 스팬 개수를 조사한다.
int GetSpanNum(int wi)
{
	// 자기 자신은 일단 포함되고 이후 위젯의 span속성이 true인 개수를 센다.
	int num = 1;
	for (int w = wi + 1; w < arWidget.num; w++) {
		if (arWidget[w].span) {
			num++;
		} else {
			break;
		}
	}

	return num;
}

// 방향과 차일드 크기에 따라 clientWidth, clientHeight를 갱신하고 레이아웃을 새로 잡는다. 
// 메인은 여기서 직접 건드리지 않으며 호출측에서 각 경우에 맞게 처리한다. 
void Relayout()
{
	int span_num;
	int totalPercent;
	int remainWidth, remainHeight;

	if (bVertical) {
		// 영역의 폭. 경계선 포함
		clientWidth = vertThick;
		// 경계선을 제외한 위젯폭
		int widgetWidth = clientWidth - bdThick * 2;

		// 타이틀바의 수직 중앙 위치에 닫기, 메뉴 버튼 배치. 
		MoveWindow(hBtnClose, clientWidth - 22, bdThick + capThick / 2 - 8, 16, 16, TRUE);
		UpdateWindow(hBtnClose);
		MoveWindow(hBtnMenu, clientWidth - 42, bdThick + capThick / 2 - 8, 16, 16, TRUE);
		UpdateWindow(hBtnMenu);

		// 모든 위젯의 위치를 다시 잡는다.
		int y = bdThick + capThick;
		for (int wi = 0; wi < arWidget.num; wi++) {
			// 숨겨진 위젯은 건너 뛴다. 병합된 위젯은 루프 내부에서 이미 배치했으므로 건너 뛴다. 
			if (arWidget[wi].visible == false || arWidget[wi].span) continue;
			// 일단 대입받은 후 사용한다. 컬렉션의 [ ] 연산자를 쓰면 디버깅중에 값을 볼 수 없다. 
			CWidget* pWidget = arWidget[wi].pWidget;
			pWidget->win_y = y;

			// 병합 개수 조사
			span_num = GetSpanNum(wi);

			// 혼자이면 자기 높이만큼 차지한다.
			if (span_num == 1) {
				MoveWindow(pWidget->win_hWnd, bdThick, pWidget->win_y, widgetWidth, pWidget->win_h, TRUE);
				// 이동 후 즉시 그려야 깜박임이 없다. 아래, 위로 순서를 바꿀 때는 무효화도 해야 바로 그려진다. 
				InvalidateRect(pWidget->win_hWnd, NULL, TRUE);
				UpdateWindow(pWidget->win_hWnd);
			} else {
				// 병합셀이면 선두 위젯(pWidget)의 높이만큼만 차지하고 나머지는 수평으로 분할 배치한다.
				// 비율의 총합을 구한다. spanRatio가 양수이면 비율, 음수이면 절대 픽셀이다.
				totalPercent = 0;
				// 위젯폭에서 중간 중간의 경계선 폭을 빼 위젯이 차지할 너비를 계산한다. 
				remainWidth = widgetWidth - (span_num - 1) * splitThick;
				for (int i = wi; i < wi + span_num; i++) {
					if (arWidget[i].spanRatio > 0) {
						// 분할을 위한 총 비율 계산
						totalPercent += arWidget[i].spanRatio;
					} else {
						// 절대폭 위젯의 너비만큼 남은 폭 제외
						remainWidth -= -arWidget[i].spanRatio;
					}
				}

				int x = bdThick;
				int w;
				// 그룹내의 위젯을 수평으로 나열한다. 
				for (int i = wi; i < wi + span_num; i++) {
					// 양수는 남는 폭의 비율만큼 폭을 차지한다.
					if (arWidget[i].spanRatio > 0) {
						w = remainWidth * arWidget[i].spanRatio / totalPercent;
					// 음수는 절대 픽셀 크기만큼 차지한다.
					} else {
						w = -arWidget[i].spanRatio;
					}
					// 마지막 위젯은 나머지를 다 차지하되 음수일 수는 없다. 비율 나누기에 의한 1 모자람 방지
					if (i == wi + span_num - 1) {
						w = clientWidth - x - bdThick;
						if (w < 0) w = 0;
					}
					MoveWindow(arWidget[i].pWidget->win_hWnd, x, pWidget->win_y, w, pWidget->win_h, TRUE);
					InvalidateRect(arWidget[i].pWidget->win_hWnd, NULL, TRUE);
					UpdateWindow(arWidget[i].pWidget->win_hWnd);

					x += (w + splitThick);
				}
			}

			y += pWidget->win_h + splitThick;
		}

		// 트레이 윈도우의 위치 지정 및 티젯 정렬
		trayWindow.win_y = y;
		UpdateTrayThick(widgetWidth);
		SetWindowPos(trayWindow.win_hWnd, NULL, bdThick, y, widgetWidth, trayThick, SWP_NOZORDER);
		InvalidateRect(trayWindow.win_hWnd, NULL, FALSE);
		UpdateWindow(trayWindow.win_hWnd);
		ArrangeTrayTidget(widgetWidth);

		// 영역의 높이 결정. 모든 위젯의 높이 + 트레이 + 아래 경계선 두께
		clientHeight = y + trayThick + bdThick;

	} else {
		// 수직 배열과 모든 논리는 같되 방향만 다르다. 
		clientHeight = horzThick;
		int widgetHeight = clientHeight - bdThick * 2;

		MoveWindow(hBtnClose, bdThick + capThick / 2 - 8, clientHeight - bdThick - 16 - 4, 16, 16, TRUE);
		UpdateWindow(hBtnClose);
		MoveWindow(hBtnMenu, bdThick + capThick / 2 - 8, clientHeight - bdThick - 16 - 4 - 20, 16, 16, TRUE);
		UpdateWindow(hBtnMenu);

		int x = bdThick + capThick;
		for (int wi = 0; wi < arWidget.num; wi++) {
			if (arWidget[wi].visible == false || arWidget[wi].span) continue;
			CWidget* pWidget = arWidget[wi].pWidget;
			pWidget->win_x = x;

			span_num = GetSpanNum(wi);

			if (span_num == 1) {
				MoveWindow(pWidget->win_hWnd, pWidget->win_x, bdThick, pWidget->win_w, widgetHeight, TRUE);
				InvalidateRect(pWidget->win_hWnd, NULL, TRUE);
				UpdateWindow(pWidget->win_hWnd);
			} else {
				totalPercent = 0;
				remainHeight = widgetHeight - (span_num - 1) * splitThick;
				for (int i = wi; i < wi + span_num; i++) {
					if (arWidget[i].spanRatio > 0) {
						totalPercent += arWidget[i].spanRatio;
					} else {
						remainHeight -= -arWidget[i].spanRatio;
					}
				}

				int y = bdThick;
				int h;
				for (int i = wi; i < wi + span_num; i++) {
					if (arWidget[i].spanRatio > 0) {
						h = remainHeight * arWidget[i].spanRatio / totalPercent;
					} else {
						h = -arWidget[i].spanRatio;
					}
					if (i == wi + span_num - 1) {
						h = clientHeight - y - bdThick;
						if (h < 0) h = 0;
					}
					MoveWindow(arWidget[i].pWidget->win_hWnd, pWidget->win_x, y, pWidget->win_w, h, TRUE);
					InvalidateRect(arWidget[i].pWidget->win_hWnd, NULL, TRUE);
					UpdateWindow(arWidget[i].pWidget->win_hWnd);

					y += (h + splitThick);
				}
			}

			x += pWidget->win_w + splitThick;
		}

		trayWindow.win_x = x;
		UpdateTrayThick(widgetHeight);
		SetWindowPos(trayWindow.win_hWnd, NULL, x, bdThick, trayThick, widgetHeight, SWP_NOZORDER);
		InvalidateRect(trayWindow.win_hWnd, NULL, FALSE);
		UpdateWindow(trayWindow.win_hWnd);
		ArrangeTrayTidget(widgetHeight);

		clientWidth = x + trayThick + bdThick;
	}

	// 메인 윈도우의 크기를 변경한다. 트레이의 높이가 바뀔 수 있어 메인도 크기가 바뀔 수 있다. 
	CustomTrace(L"clientWidth = %d, clientHeight=%d\r\n", clientWidth, clientHeight);
	SetWindowPos(mainWindow.win_hWnd, NULL, 0, 0, clientWidth, clientHeight, SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);
	InvalidateRect(mainWindow.win_hWnd, NULL, FALSE);
}

// 트레이의 두께를 결정한다. 인수로 경계선을 제외한 위젯 두께를 전달한다.
// 수평, 수직에 상관 없이 논리는 같되 Thick이 폭이냐 높이냐만 다르다.
void UpdateTrayThick(int widgetThick)
{
	// 한줄에 들어가는 티젯 개수. 위젯폭에서 왼쪽의 갭 빼고 폭+갭으로 나눈 수.
	// 예를 들어 위젯폭이 100일 때 첫 갭인 4를 뺀 96에 폭+갭인 28이 몇 개나 들어가는지 본다. 이 경우는 3개이다.
	int perLine = (widgetThick - tidgetGap) / (tidgetSize + tidgetGap);
	// 보이는 티젯의 개수를 센다.
	int num = 0;
	for (int ti = 0; ti < arTidget.num; ti++) {
		if (arTidget[ti].visible == false) continue;
		num++;
	}
	// 개수에 따른 줄 수. 1, 2, 3개까지는 1줄이고 4, 5, 6은 두 줄이다. 게시물 페이징하는 수식과 같다.
	int trayLine = (num - 1) / perLine + 1;
	// 트레이의 두께 = 위쪽 갭에서 폭+갭 높이와 줄 수를 곱한 값이다.
	trayThick = tidgetGap + (tidgetSize + tidgetGap) * trayLine;

	CustomTrace(L"UpdateTrayThick. trayThick = %d\r\n", trayThick);
}

// 트레이의 티젯을 배치한다. 수직은 좌우에서 우로, 위에서 아래로 배치하고 수평은 위에서 아래로, 좌에서 우로 배치한다.
// 방향만 다를 뿐 배치 논리는 동일하다. 
void ArrangeTrayTidget(int widgetThick)
{
	int x, y;

	if (bVertical) {
		x = tidgetGap;
		y = tidgetGap;
		for (int ti = 0; ti < arTidget.num; ti++) {
			if (arTidget[ti].visible == false) continue;
			// 티젯 오른쪽이 트레이 오른쪽에 닿았으면 개행한다. 제일 끝 아이콘의 갭까지도 포함해야 한다.
			if (x + tidgetSize + tidgetGap > widgetThick) {
				x = tidgetGap;
				y += (tidgetSize + tidgetGap);
			}
			MoveWindow(arTidget[ti].pTidget->win_hWnd, x, y, tidgetSize, tidgetSize, TRUE);
			UpdateWindow(arTidget[ti].pTidget->win_hWnd);
			x += (tidgetSize + tidgetGap);
		}
	} else {
		y = tidgetGap;
		x = tidgetGap;
		for (int ti = 0; ti < arTidget.num; ti++) {
			if (arTidget[ti].visible == false) continue;
			if (y + tidgetSize + tidgetGap > widgetThick) {
				y = tidgetGap;
				x += (tidgetSize + tidgetGap);
			}
			MoveWindow(arTidget[ti].pTidget->win_hWnd, x, y, tidgetSize, tidgetSize, TRUE);
			UpdateWindow(arTidget[ti].pTidget->win_hWnd);
			y += (tidgetSize + tidgetGap);
		}
	}
}

// 몇번째 위젯 스플릿터 아래에 있는지 조사한다. 아니면 -1 리턴
int FindSplitVirt(int cy)
{
	int y = bdThick + capThick;
	for (int k = 0; k < arWidget.num; k++) {
		if (arWidget[k].visible == false || arWidget[k].span) continue;
		y += arWidget[k].pWidget->win_h;
		if (cy < y) return -1;
		// 스플릿터 사이에 있으면 수직 조정 커서
		if (cy >= y && cy <= y + splitThick) {
			return k;
		}
		y += splitThick;
	}

	return -1;
}

int FindSplitHorz(int cx)
{
	int x = bdThick + capThick;
	for (int k = 0; k < arWidget.num; k++) {
		if (arWidget[k].visible == false || arWidget[k].span) continue;
		x += arWidget[k].pWidget->win_w;
		if (cx < x) return -1;
		if (cx >= x && cx <= x + splitThick) {
			return k;
		}
		x += splitThick;
	}

	return -1;
}

bool IsInCaption(int cx, int cy)
{
	if (bVertical) {
		if (cy <= capThick) return true;
	} else {
		if (cx <= capThick) return true;
	}
	return false;
}

INT_PTR CALLBACK SelWidgetDlgProc(HWND hDlg, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	TCHAR name[64];
	TCHAR id[32];
	int idx;
	int idx_dget;
	int idx_use;
	int cidx;
	int count;
	TCHAR* p;
	int value;
	static bool editByCode = false;
	LPMEASUREITEMSTRUCT lpmis;
	LPDRAWITEMSTRUCT lpdis;
	HBRUSH bkBrush, Brush, OldBrush;
	static COLORREF arWidgetBack[24] = { 0, };

	idx_use = SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_GETCURSEL, 0, 0);

	switch (iMessage) {
	case WM_INITDIALOG:
		MovePopupToSide(hDlg, mainWindow.win_x, mainWindow.win_y);

		// 표준 색상표에서 밝은 파스텔 계통으로 선별함. 
		if (arWidgetBack[0] == 0) {
			arWidgetBack[0] = GetColorFromName(L"Pink");
			arWidgetBack[1] = GetColorFromName(L"Khaki");
			arWidgetBack[2] = GetColorFromName(L"LightYellow");
			arWidgetBack[3] = GetColorFromName(L"PapayaWhip");
			arWidgetBack[4] = GetColorFromName(L"Moccasin");
			arWidgetBack[5] = GetColorFromName(L"PaleTurquoise");
			arWidgetBack[6] = GetColorFromName(L"LightBlue");
			arWidgetBack[7] = GetColorFromName(L"LightSkyBlue");
			arWidgetBack[8] = GetColorFromName(L"SpringGreen");
			arWidgetBack[9] = GetColorFromName(L"MediumAquaMarine");
			arWidgetBack[10] = GetColorFromName(L"Turquoise");
			arWidgetBack[11] = GetColorFromName(L"Lavender");
			arWidgetBack[12] = GetColorFromName(L"Thistle");
			arWidgetBack[13] = GetColorFromName(L"Violet");
			arWidgetBack[14] = GetColorFromName(L"HoneyDew");
			arWidgetBack[15] = GetColorFromName(L"Beige");
			arWidgetBack[16] = GetColorFromName(L"Ivory");
			arWidgetBack[17] = GetColorFromName(L"CornSilk");
			arWidgetBack[18] = GetColorFromName(L"Bisque");
			arWidgetBack[19] = GetColorFromName(L"Wheat");
			arWidgetBack[20] = GetColorFromName(L"Gainsboro");
			arWidgetBack[21] = GetColorFromName(L"LightGrey");
			arWidgetBack[22] = GetColorFromName(L"Silver");
			arWidgetBack[23] = GetColorFromName(L"DarkGray");
		}

		// 색상수만큼 콤보 박스에 채운다.
		for (int i = 0; i < Count(arWidgetBack); i++) {
			SendDlgItemMessage(hDlg, IDC_CBBACKCOLOR, CB_ADDSTRING, 0, 0);
		}

		// 사용 가능한 위젯 목록 출력. 
		for (int wi = 0; wi < Count(arDget); wi++) {
			if (arDget[wi].isWidget == false) continue;
			wsprintf(name, L"%s", arDget[wi].name);
			if (arDget[wi].allowMulti == false) {
				lstrcat(name, L"(멀티안됨)");
			}
			// 이름과 arDget 배열상의 첨자를 저장한다. 항목 데이터에 포인터를 저장하는 것은 위험하다. 
			// atDget의 배열상 첨자와 리스트 박스의 첨자는 같지 않아 따로 저장해야 한다. 
			idx = SendDlgItemMessage(hDlg, IDC_LIWIDGET, LB_ADDSTRING, 0, (LPARAM)name);
			SendDlgItemMessage(hDlg, IDC_LIWIDGET, LB_SETITEMDATA, idx, (LPARAM)wi);
		}

		// 현재 사용중인 위젯 목록 출력. 이름과 실제 id를 같이 저장. 인스턴스간의 구분을 위한 id가 필요하다. 
		// 첨자는 삭제에 의해 무효해져서 안되며 항목 데이터에 포인터 할당해서 저장하는 것은 번거롭다. 
		for (int wi = 0; wi < arWidget.num; wi++) {
			wsprintf(name, L"%s-%s", arWidget[wi].name, arWidget[wi].id);
			SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_ADDSTRING, 0, (LPARAM)name);
		}

		// 스핀 컨트롤의 범위 설정
		SendDlgItemMessage(hDlg, IDC_SPINHEIGHT, UDM_SETRANGE, 0, MAKELPARAM(widgetMaxThick, widgetMinThick));
		SendDlgItemMessage(hDlg, IDC_SPINWIDTH, UDM_SETRANGE, 0, MAKELPARAM(widgetMaxThick, widgetMinThick));
		SendDlgItemMessage(hDlg, IDC_SPINSPAN, UDM_SETRANGE, 0, MAKELPARAM(widgetMaxThick, 1));

		// 두 리스트 박스 모두 첫 번째 항목을 선택해 준다. 
		SendDlgItemMessage(hDlg, IDC_LIWIDGET, LB_SETCURSEL, 0, 0);
		SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_SETCURSEL, 0, 0);
		SendMessage(hDlg, WM_COMMAND, MAKEWPARAM(IDC_LIUSEWIDGET, LBN_SELCHANGE), NULL);
		return TRUE;
	case WM_MEASUREITEM:
		lpmis = (LPMEASUREITEMSTRUCT)lParam;
		lpmis->itemHeight = 18;
		return TRUE;
	case WM_DRAWITEM:
		if (wParam == IDC_CBBACKCOLOR) {
			lpdis = (LPDRAWITEMSTRUCT)lParam;

			// 선택 배경 그림
			if (lpdis->itemState & ODS_SELECTED) {
				bkBrush = CreateSolidBrush(RGB(0, 0, 255));
			} else {
				bkBrush = CreateSolidBrush(RGB(255, 255, 255));
			}
			FillRect(lpdis->hDC, &lpdis->rcItem, bkBrush);
			DeleteObject(bkBrush);

			// 색상 그림
			Brush = CreateSolidBrush(arWidgetBack[lpdis->itemID]);
			OldBrush = (HBRUSH)SelectObject(lpdis->hDC, Brush);
			Rectangle(lpdis->hDC, lpdis->rcItem.left + 2, lpdis->rcItem.top + 1,
				lpdis->rcItem.right - 2, lpdis->rcItem.bottom - 1);
			SelectObject(lpdis->hDC, OldBrush);
			DeleteObject(Brush);
			return TRUE;
		}
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDC_BTNADDWIDGET:
			idx_dget = SendDlgItemMessage(hDlg, IDC_LIWIDGET, LB_GETCURSEL, 0, 0);
			if (idx_dget != -1) {
				// 추가할 위젯의 정보를 구한다. 
				SendDlgItemMessage(hDlg, IDC_LIWIDGET, LB_GETTEXT, idx_dget, (LPARAM)name);
				idx = SendDlgItemMessage(hDlg, IDC_LIWIDGET, LB_GETITEMDATA, idx_dget, 0);
				lstrcpy(id, arDget[idx].initId);

				// 복수 비허용일 경우 이미 사용중인지 조사한다. 이때는 숨겨진 위젯까지 포함한다.
				if (arDget[idx].allowMulti == false) {
					bool bFind = false;
					for (int wi = 0; wi < arWidget.num; wi++) {
						if (lstrcmp(arWidget[wi].initId, id) == 0) {
							bFind = true;
							break;
						}
					}
					if (bFind) {
						MessageBox(hDlg, L"이 위젯은 여러 개 사용할 수 없거나 그럴 필요가 없습니다.", L"알림", MB_OK);
						return TRUE;
					}
				}

				// 새 위젯을 추가한다. 
				sWidgetInfo widget;
				CopyWidgetInfo(id, &widget); arWidget.Append(widget);
				int w = arWidget.num - 1;

				// 새 위젯 생성.
				CreateWidget(w, mainWindow.win_hWnd);

				// 새 위젯을 사용 위젯 리스트 박스에 추가
				wsprintf(name, L"%s-%s", arWidget[w].name, arWidget[w].id);
				idx = SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_ADDSTRING, 0, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_SETCURSEL, idx, 0);
				SendMessage(hDlg, WM_COMMAND, MAKEWPARAM(IDC_LIUSEWIDGET, LBN_SELCHANGE), NULL);

				// 새로 생성한 위젯을 화면에 바로 출력한다. 
				Relayout();
			}
			return TRUE;
		case IDC_BTNREMOVEWIDGET:
			if (idx_use != -1) {
				if (MessageBox(hDlg, L"위젯 제거시 모든 정보도 같이 삭제됩니다. 제거하시겠습니까?", L"확인", MB_YESNO) == IDNO) {
					return TRUE;
				}
				// 삭제할 위젯의 id를 구해 놓고 삭제한다.
				memset(name, 0, sizeof(name));
				SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_GETTEXT, idx_use, (LPARAM)name);
				p = _tcschr(name, '-') + 1;
				lstrcpy(id, p);
				SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_DELETESTRING, idx_use, 0);

				// 위젯 윈도우 파괴. 
				DestroyWindow(arWidget[idx_use].pWidget->win_hWnd);
				// 위젯 객체도 제거
				delete arWidget[idx_use].pWidget;
				// 위젯 삭제
				arWidget.Delete(idx_use);

				// 삭제되는 위젯의 이전 정보도 모두 제거. DestroyWindow에서 정보를 기록하므로 윈도우 파괴 후 제거해야 함.
				// 이전 정보를 제거하지 않으면 다음 추가하는 위젯이 id를 재활용하므로 쓰레기 정보를 읽게 된다.
				setting.deleteSection(id);

				Relayout();
				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_SETCURSEL, idx_use, 0);
			}

			return TRUE;
		case IDC_BTNUPWIDGET:
			if (idx_use != -1 && idx_use != 0) {
				// 아래, 위 위젯 교환. 구조체끼리의 교환이어서 대입만 하면 된다. 
				sWidgetInfo t = arWidget[idx_use];
				arWidget[idx_use] = arWidget[idx_use - 1];
				arWidget[idx_use - 1] = t;

				// 리스트 박스의 현재 항목 백업 후 삭제
				SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_GETTEXT, idx_use, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_DELETESTRING, idx_use, 0);

				// 바로 위에 다시 삽입하고 포커스를 준다. 
				idx = SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_INSERTSTRING, idx_use - 1, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_SETCURSEL, idx, 0);

				Relayout();
			}
			return TRUE;
		case IDC_BTNDOWNWIDGET:
			count = SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_GETCOUNT, 0, 0);
			if (idx_use != -1 && idx_use != count - 1) {
				// 아래, 위 위젯 교환. 구조체끼리의 교환이어서 대입만 하면 된다. 
				sWidgetInfo t = arWidget[idx_use];
				arWidget[idx_use] = arWidget[idx_use + 1];
				arWidget[idx_use + 1] = t;

				// 리스트 박스의 현재 항목 백업 후 삭제
				SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_GETTEXT, idx_use, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_DELETESTRING, idx_use, 0);

				// 바로 위에 다시 삽입하고 포커스를 준다. 
				idx = SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_INSERTSTRING, idx_use + 1, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_SETCURSEL, idx, 0);

				Relayout();
			}
			return TRUE;
		case IDCANCEL:
			EndDialog(hDlg, IDCANCEL);
			return TRUE;
		case IDC_LIUSEWIDGET:
			switch (HIWORD(wParam)) {
			case LBN_SELCHANGE:
				if (idx_use != -1) {
					CheckDlgButton(hDlg, IDC_CHKHIDEWIDGET, arWidget[idx_use].visible ? BST_UNCHECKED : BST_CHECKED);
					editByCode = true;
					SetDlgItemInt(hDlg, IDC_EDHEIGHT, arWidget[idx_use].height, FALSE);
					SetDlgItemInt(hDlg, IDC_EDWIDTH, arWidget[idx_use].width, FALSE);

					// 양수면 상대 비율, 음수면 절대 픽셀
					CheckDlgButton(hDlg, IDC_CHKSPAN, arWidget[idx_use].span ? BST_CHECKED : BST_UNCHECKED);
					CheckRadioButton(hDlg, IDC_RASPANRATIO, IDC_RASPANPIXEL,
						arWidget[idx_use].spanRatio > 0 ? IDC_RASPANRATIO : IDC_RASPANPIXEL);
					SetDlgItemInt(hDlg, IDC_EDSPAN, ABS(arWidget[idx_use].spanRatio), FALSE);
					editByCode = false;

					// 위젯의 색상을 선택한다. 색상표에 없으면 0번을 선택한다.
					cidx = 0;
					for (int i = 0; i < Count(arWidgetBack); i++) {
						if (arWidget[idx_use].backColor == arWidgetBack[i]) {
							cidx = i;
							break;
						}
					}
					SendDlgItemMessage(hDlg, IDC_CBBACKCOLOR, CB_SETCURSEL, cidx, 0);
				}
				return TRUE;
			}
			break;
		case IDC_CHKHIDEWIDGET:
			if (idx_use != -1) {
				arWidget[idx_use].visible = (IsDlgButtonChecked(hDlg, IDC_CHKHIDEWIDGET) == BST_UNCHECKED);
				ShowWindow(arWidget[idx_use].pWidget->win_hWnd, arWidget[idx_use].visible ? SW_SHOW : SW_HIDE);
				// 위젯을 숨기면 분할에는 참여할 수 없다. 만약 선두 위젯을 숨기면 그룹 전체를 건너뛴다. 
				if (arWidget[idx_use].visible == false) {
					arWidget[idx_use].span = false;
					CheckDlgButton(hDlg, IDC_CHKSPAN, BST_UNCHECKED);
				}
				Relayout();
			}
			return TRUE;
		case IDC_EDHEIGHT:
			switch(HIWORD(wParam)) {
			case EN_CHANGE:
				// 코드에서 값을 변경할 때는 이벤트 처리할 필요 없다. 
				if (editByCode) return TRUE;
				value = GetDlgItemInt(hDlg, IDC_EDHEIGHT, NULL, FALSE);
				value = min(max(value, widgetMinThick), widgetMaxThick);
				if (idx_use != -1) {
					// 둘 중 어떤게 맞는지 모르겠네. 일단은 win_h가 맞음
					arWidget[idx_use].height = value;
					arWidget[idx_use].pWidget->win_h = value;
					if (bVertical) {
						SetTimer(hDlg, 1, 800, NULL);
					}
					return TRUE;
				}
			}
			break;
		case IDC_EDWIDTH:
			switch (HIWORD(wParam)) {
			case EN_CHANGE:
				if (editByCode) return TRUE;
				value = GetDlgItemInt(hDlg, IDC_EDWIDTH, NULL, FALSE);
				value = min(max(value, widgetMinThick), widgetMaxThick);
				if (idx_use != -1) {
					arWidget[idx_use].width = value;
					arWidget[idx_use].pWidget->win_w = value;
					if (bVertical == false) {
						SetTimer(hDlg, 1, 800, NULL);
					}
					return TRUE;
				}
			}
			break;
		case IDC_CHKSPAN:
			if (idx_use != -1) {
				arWidget[idx_use].span = (IsDlgButtonChecked(hDlg, IDC_CHKSPAN) == BST_CHECKED);
				SetTimer(hDlg, 1, 800, NULL);
			}
			return TRUE;
		case IDC_EDSPAN:
			switch (HIWORD(wParam)) {
			case EN_CHANGE:
				if (editByCode) return TRUE;
				value = GetDlgItemInt(hDlg, IDC_EDSPAN, NULL, FALSE);
				if (idx_use != -1) {
					if (IsDlgButtonChecked(hDlg, IDC_RASPANRATIO) == BST_CHECKED) {
						arWidget[idx_use].spanRatio = value;
					} else {
						arWidget[idx_use].spanRatio = -value;
					}
					SetTimer(hDlg, 1, 800, NULL);
				}
				return TRUE;
			}
			break;
		case IDC_RASPANRATIO:
			if (idx_use != -1) {
				value = GetDlgItemInt(hDlg, IDC_EDSPAN, NULL, FALSE);
				arWidget[idx_use].spanRatio = value;
				SetTimer(hDlg, 1, 800, NULL);
			}
			return TRUE;
		case IDC_RASPANPIXEL:
			if (idx_use != -1) {
				value = GetDlgItemInt(hDlg, IDC_EDSPAN, NULL, FALSE);
				arWidget[idx_use].spanRatio = -value;
				SetTimer(hDlg, 1, 800, NULL);
			}
			return TRUE;
		case IDC_CBBACKCOLOR:
			switch (HIWORD(wParam)) {
			case CBN_SELCHANGE:
				idx = SendDlgItemMessage(hDlg, LOWORD(wParam), CB_GETCURSEL, 0, 0);
				idx_use = SendDlgItemMessage(hDlg, IDC_LIUSEWIDGET, LB_GETCURSEL, 0, 0);
				if (idx_use != -1) {
					arWidget[idx_use].backColor = arWidgetBack[idx];
					if (arWidget[idx_use].pWidget->win_backBrush != NULL) {
						DeleteObject(arWidget[idx_use].pWidget->win_backBrush);
					}
					arWidget[idx_use].pWidget->win_backBrush = CreateSolidBrush(arWidget[idx_use].backColor);
					InvalidateRect(arWidget[idx_use].pWidget->win_hWnd, NULL, TRUE);
				}
				break;
			}
		}	// end WM_COMMAND
		break;
	// 일회용 타이머. 에디트 편집시 즉시 처리하지 말고 0.8초간 시차를 두고 레이아웃에 반영한다.
	case WM_TIMER:
		if (wParam == 1) {
			KillTimer(hDlg, 1);
			Relayout();
		}
	}
	return FALSE;
}

INT_PTR CALLBACK SelTidgetDlgProc(HWND hDlg, UINT iMessage, WPARAM wParam, LPARAM lParam)
{
	TCHAR name[64];
	TCHAR id[32];
	int idx;
	int idx_dget;
	int idx_use;
	TCHAR* p;
	int count;

	idx_use = SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_GETCURSEL, 0, 0);

	switch (iMessage) {
	case WM_INITDIALOG:
		MovePopupToSide(hDlg, mainWindow.win_x, mainWindow.win_y);

		// 사용 가능한 티젯 목록 출력. 
		for (int ti = 0; ti < Count(arDget); ti++) {
			if (arDget[ti].isWidget == true) continue;
			wsprintf(name, L"%s", arDget[ti].name);
			if (arDget[ti].allowMulti == false) {
				lstrcat(name, L"(멀티안됨)");
			}
			idx = SendDlgItemMessage(hDlg, IDC_LITIDGET, LB_ADDSTRING, 0, (LPARAM)name);
			SendDlgItemMessage(hDlg, IDC_LITIDGET, LB_SETITEMDATA, idx, (LPARAM)ti);
		}

		for (int ti = 0; ti < arTidget.num; ti++) {
			wsprintf(name, L"%s-%s", arTidget[ti].name, arTidget[ti].id);
			SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_ADDSTRING, 0, (LPARAM)name);
		}

		// 두 리스트 박스 모두 첫 번째 항목을 선택해 준다. 
		SendDlgItemMessage(hDlg, IDC_LITIDGET, LB_SETCURSEL, 0, 0);
		SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_SETCURSEL, 0, 0);
		SendMessage(hDlg, WM_COMMAND, MAKEWPARAM(IDC_LIUSETIDGET, LBN_SELCHANGE), NULL);
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDC_BTNADDTIDGET:
			idx_dget = SendDlgItemMessage(hDlg, IDC_LITIDGET, LB_GETCURSEL, 0, 0);
			if (idx_dget != -1) {
				// 추가할 위젯의 정보를 구한다. 
				SendDlgItemMessage(hDlg, IDC_LITIDGET, LB_GETTEXT, idx_dget, (LPARAM)name);
				idx = SendDlgItemMessage(hDlg, IDC_LITIDGET, LB_GETITEMDATA, idx_dget, 0);
				lstrcpy(id, arDget[idx].initId);

				// 복수 비허용일 경우 이미 사용중인지 조사한다. 이때는 숨겨진 위젯까지 포함한다.
				if (arDget[idx].allowMulti == false) {
					bool bFind = false;
					for (int ti = 0; ti < arTidget.num; ti++) {
						if (lstrcmp(arTidget[ti].initId, id) == 0) {
							bFind = true;
							break;
						}
					}
					if (bFind) {
						MessageBox(hDlg, L"이 티젯은 여러 개 사용할 수 없거나 그럴 필요가 없습니다.", L"알림", MB_OK);
						return TRUE;
					}
				}

				// 새 티젯을 추가한다. 
				sTidgetInfo Tidget;
				CopyTidgetInfo(id, &Tidget); arTidget.Append(Tidget);
				int t = arTidget.num - 1;

				// 새 위젯 생성.
				CreateTidget(t, trayWindow.win_hWnd);

				// 새 티젯을 사용 티젯 리스트 박스에 추가
				wsprintf(name, L"%s-%s", arTidget[t].name, arTidget[t].id);
				idx = SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_ADDSTRING, 0, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_SETCURSEL, idx, 0);
				SendMessage(hDlg, WM_COMMAND, MAKEWPARAM(IDC_LIUSETIDGET, LBN_SELCHANGE), NULL);

				// 새로 생성한 티젯을 화면에 바로 출력한다. 
				Relayout();
			}
			return TRUE;
		case IDC_BTNREMOVETIDGET:
			if (idx_use != -1) {
				if (MessageBox(hDlg, L"티젯 제거시 모든 정보도 같이 삭제됩니다. 제거하시겠습니까?", L"확인", MB_YESNO) == IDNO) {
					return TRUE;
				}
				memset(name, 0, sizeof(name));
				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_GETTEXT, idx_use, (LPARAM)name);
				p = _tcschr(name, '-') + 1;
				lstrcpy(id, p);
				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_DELETESTRING, idx_use, 0);

				DestroyWindow(arTidget[idx_use].pTidget->win_hWnd);
				delete arTidget[idx_use].pTidget;
				arTidget.Delete(idx_use);

				setting.deleteSection(id);
				Relayout();
				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_SETCURSEL, idx_use, 0);
			}

			return TRUE;
		case IDC_BTNUPTIDGET:
			if (idx_use != -1 && idx_use != 0) {
				sTidgetInfo t = arTidget[idx_use];
				arTidget[idx_use] = arTidget[idx_use - 1];
				arTidget[idx_use - 1] = t;

				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_GETTEXT, idx_use, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_DELETESTRING, idx_use, 0);

				idx = SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_INSERTSTRING, idx_use - 1, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_SETCURSEL, idx, 0);

				Relayout();
			}
			return TRUE;
		case IDC_BTNDOWNTIDGET:
			count = SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_GETCOUNT, 0, 0);
			if (idx_use != -1 && idx_use != count - 1) {
				sTidgetInfo t = arTidget[idx_use];
				arTidget[idx_use] = arTidget[idx_use + 1];
				arTidget[idx_use + 1] = t;

				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_GETTEXT, idx_use, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_DELETESTRING, idx_use, 0);

				idx = SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_INSERTSTRING, idx_use + 1, (LPARAM)name);
				SendDlgItemMessage(hDlg, IDC_LIUSETIDGET, LB_SETCURSEL, idx, 0);

				Relayout();
			}
			return TRUE;
		case IDCANCEL:
			EndDialog(hDlg, IDCANCEL);
			return TRUE;
		case IDC_LIUSETIDGET:
			switch (HIWORD(wParam)) {
			case LBN_SELCHANGE:
				if (idx_use != -1) {
					CheckDlgButton(hDlg, IDC_CHKHIDETIDGET, arTidget[idx_use].visible ? BST_UNCHECKED : BST_CHECKED);
				}
				return TRUE;
			}
			break;
		case IDC_CHKHIDETIDGET:
			if (idx_use != -1) {
				arTidget[idx_use].visible = (IsDlgButtonChecked(hDlg, IDC_CHKHIDETIDGET) == BST_UNCHECKED);
				ShowWindow(arTidget[idx_use].pTidget->win_hWnd, arTidget[idx_use].visible ? SW_SHOW : SW_HIDE);
				Relayout();
			}
			return TRUE;
		}	// end WM_COMMAND
		break;
	}
	return FALSE;
}
